SystemOrganization addCategory: #'Math-KDTree'!

Array variableSubclass: #NNStore
	instanceVariableNames: 'maxDistance lastUsedIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Math-KDTree'!

!NNStore commentStamp: 'WernerKassens 5/5/2011 14:40' prior: 0!
NNStore stores neighbours together with distances and uses the distances for sorting. 

can easily be subclassed if one needs an Array sorted depending on some outside data. see StupidNN for a simple example and comment on "withAll:" or just use "newFrom:".!

!NNStore class methodsFor: 'instance creation' stamp: 'WernerKassens 5/3/2011 21:54'!
new: anInt
^((super new: anInt) atAllPut: {Float infinity. nil}  )initialize  ! !

!NNStore class methodsFor: 'instance creation' stamp: 'WernerKassens 3/6/2014 14:44'!
newFrom: aCollectionWithSortingIndex
"example:
( NNStore newFrom:#(#(3 1.0) #(0 '2')) ) data.   #('2' 1.0)"
^((super newFrom: aCollectionWithSortingIndex )lastUsedIndex: aCollectionWithSortingIndex size )sort  ! !

!NNStore class methodsFor: 'instance creation' stamp: 'WernerKassens 3/6/2014 14:44'!
withAll: aCollection
"this is the peferable form of instance creation for subclasses since freeIndex is correctly initialized this way. then 'sortFor:' must be overwritten and called at least once (!!); an example is StupidNN. the other possibility is to use 'new:' and to fill NNStore subclass with 'add:' . the advantage is, that NNStore has not to be filled completely this way and the sorting logic can be applied outside of the subclass without sortFor: (see also comment on add:). 
you cant enter the sorting index with the method withAll: . if you want to do that, use 'newFrom:' instead"
^(super withAll: (aCollection collect: [:coll| {nil. coll}]))lastUsedIndex: aCollection size  ! !

!NNStore methodsFor: 'comparing' stamp: 'WernerKassens 3/9/2014 09:59'!
= aNNStore
self == aNNStore ifTrue: [^true].
((self species =aNNStore species ) and:[(maxDistance =aNNStore maxDistance)and:[self size=aNNStore size ] ])
		ifFalse: [^false].
^ self completeData  =aNNStore completeData 
! !

!NNStore methodsFor: 'adding' stamp: 'WernerKassens 3/6/2014 14:40'!
add: distAndNeighbour
"distAndNeighbour is a SequenceableCollection of the sorting index and the data value.
 dist, the sorting index, has to be a number. distAndNeighbour will only be added if dist is as good or better (smaller) than maxDistance. if NNStore is full, further data will be added at the last position and the last data, which must be worse than distAndNeighbour, will be overwritten."
	(maxDistance > distAndNeighbour first) ifFalse: [^self]. 
	self isFull ifFalse: [lastUsedIndex :=lastUsedIndex +1] .
	self at:  lastUsedIndex  put: distAndNeighbour.
	self isFull ifTrue: [ self sort ].! !

!NNStore methodsFor: 'accessing' stamp: 'WernerKassens 3/9/2014 08:51'!
completeData
"return the sorted data with distance (at first place)"
|ar |
ar :=(Array new:lastUsedIndex) replaceFrom: 1 to:  lastUsedIndex with: self startingAt:  1 .
^self isFull ifTrue: [ar] ifFalse: [ar sort: [: a :b| a first < b first ] ]
! !

!NNStore methodsFor: 'copying' stamp: 'WernerKassens 3/6/2014 14:41'!
copyEmpty
	^ self species new: lastUsedIndex ! !

!NNStore methodsFor: 'accessing' stamp: 'wk 2/28/2014 12:15'!
data
"return the sorted data"
^self completeData  collect: [:e|e at:2] ! !

!NNStore methodsFor: 'comparing' stamp: 'WernerKassens 3/9/2014 08:54'!
hash
^self class hash bitXor: (self size hash bitXor:(maxDistance hash bitXor: self completeData hash)) ! !

!NNStore methodsFor: 'initialize-release' stamp: 'wk 3/9/2014 12:46'!
initialize 
"can also be used for resetting as eg in testEqual2"
lastUsedIndex :=0.
maxDistance := ((self size = 0) ifTrue: [ 0.0 - Float infinity ]ifFalse: [ Float infinity ]). "(self size = 0) ifTrue: is necessary here, ifEmpty wouldnt work"
! !

!NNStore methodsFor: 'testing' stamp: 'WernerKassens 3/6/2014 14:43'!
isEmpty 
^lastUsedIndex =0! !

!NNStore methodsFor: 'testing' stamp: 'WernerKassens 3/6/2014 14:43'!
isFull
^lastUsedIndex =self size ! !

!NNStore methodsFor: 'private' stamp: 'wk 3/9/2014 12:50'!
lastUsedIndex: anInt
"method is only necessary for correct instance creation. not directly usable for deleting data!!"
lastUsedIndex :=anInt .! !

!NNStore methodsFor: 'accessing' stamp: 'wk 2/28/2014 11:43'!
maxDistance 
"the maximum sorting index, if NNStore is full; if not then Infinity"
^maxDistance ! !

!NNStore methodsFor: 'copying' stamp: 'WernerKassens 3/6/2014 14:37'!
postCopy 
super postCopy .
lastUsedIndex :=lastUsedIndex copy.
maxDistance :=maxDistance copy.

! !

!NNStore methodsFor: 'private' stamp: 'WernerKassens 3/3/2014 18:33'!
sort
self ifEmpty: [^self].
self sort: [:a :b| a first < b first ] .
 maxDistance := self last first .! !

!NNStore methodsFor: 'sorting' stamp: 'WernerKassens 3/4/2014 16:34'!
sortFor: whatever
"should be overwritten by subclasses"
"
canonical form:
self do:[:e|e at: 1 put: somethingDependingOnWhatever ].
self sort.
"
"sortFor: should not be called before the NNStore is full!!"
self do:[:e|e at: 1 put: (e at: 2) ].
self sort .! !

NNStore variableSubclass: #StupidNN
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Math-KDTree'!

!StupidNN commentStamp: 'wk 2/23/2014 13:32' prior: 0!
StupidNN is a naive nearest neighbour search. KDTree is much faster though, if you have to do several searches. StupidNN exists only for the KDTreeTests (and as an example, how simple it is to subclass NNStore) .!

!StupidNN methodsFor: 'evaluating' stamp: 'WernerKassens 5/4/2011 13:38'!
nnSearch: aCollection i:anInt
"search for i nearest neighbours of vector aCollection "
|d|
self sortFor: aCollection   .
d :=self data .
anInt =1 ifTrue:[^d first ].
^d size >anInt ifTrue: [d copyFrom: 1 to: anInt]ifFalse: [d]  
  ! !

!StupidNN methodsFor: 'sorting' stamp: 'WernerKassens 3/4/2014 15:16'!
sortFor: aCollection
"sort nearest neighbours of aCollection "
|i|
self do:[:e| i :=aCollection -(e at:2) . e at: 1 put: (i*i)sum ];
	sort .! !

Object subclass: #KDTree
	instanceVariableNames: 'left right dim value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Math-KDTree'!

!KDTree commentStamp: 'WernerKassens 5/5/2011 15:45' prior: 0!
KDTree is a space partitioning data structure to store points of that space with "KDTree withAll: aCollectionOfVectors", where a vector can be any collection of numbers that understands "at:". you can then find nearest neighbours of aVector with "nnSearch: aVector i: numberOfNearestNeighbours".   !

KDTree subclass: #IndexedKDTree
	instanceVariableNames: 'index'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Math-KDTree'!

!IndexedKDTree commentStamp: '<historical>' prior: 0!
IndexedKDTree returns the indices of the nearest neighbours instead of the nearest neighbours itself. additionally it returns the squared distances between the vector and its neighbours.!

!IndexedKDTree class methodsFor: 'instance creation' stamp: 'WernerKassens 3/6/2014 16:00'!
withAll: aCollectionOfCollections
"make a KDTree from a SequenceableCollection of SequenceableCollections ,which is an ordered Collection of points in a n-dimensional space. the nearest neighbour search then returns the indices of the neighbours"
|c|
c :=aCollectionOfCollections withIndexCollect: [:v :i| {i.v.}]  .
^super withAll: c.  ! !

!IndexedKDTree methodsFor: 'private' stamp: 'WernerKassens 3/3/2014 15:21'!
add: aDistance to: aNNStore
aNNStore add: {aDistance . {index . value}} .! !

!IndexedKDTree methodsFor: 'private' stamp: 'wk 6/10/2013 00:47'!
adjustValue 

index :=value first. 
value :=value at: 2.! !

!IndexedKDTree methodsFor: 'evaluating' stamp: 'WernerKassens 3/5/2014 10:30'!
nnSearch: aSequenceableCollection i: anInt
"search for i nearest neighbours of vector aSequenceableCollection and return the indices; and squared distances (if i>1, at second position)"
|n|
n :=NNStore new: anInt .
self nnSearch: aSequenceableCollection near: n .
^anInt =1 ifTrue: [n data first first ] ifFalse:   [n completeData collect: [:i | {(i at:2)first . i first}]  ]  ! !

!IndexedKDTree methodsFor: 'initialize' stamp: 'WernerKassens 3/5/2014 10:29'!
sort: aSequenceableCollection
^ aSequenceableCollection sort:  [:a :b| ((a at: 2)  at: dim) < ((b at: 2) at: dim)] ! !

!KDTree class methodsFor: 'instance creation' stamp: 'wk 3/9/2014 12:40'!
withAll: aCollectionOfCollections
"make a KDTree from a Collection of SequenceableCollections ,which is a Collection of points in a n-dimensional space"
(aCollectionOfCollections allSatisfy: [:c|c isCollection ])ifFalse: 			"isSequenceable does not work here in a simple way"
	[self error: 'KDTree withAll: expects a collection of collections']. 
aCollectionOfCollections first ifEmpty: [self error: 'KDTree withAll: error'  ].
^self new addData: aCollectionOfCollections asOrderedCollection at:1.     ! !

!KDTree methodsFor: 'private' stamp: 'WernerKassens 3/3/2014 15:20'!
add: aDistance to: aNNStore
aNNStore add: {aDistance . value} .! !

!KDTree methodsFor: 'initialize' stamp: 'WernerKassens 3/5/2014 10:29'!
addData: aSequenceableCollection at: dimension
|sorted cut nextDimension|
dim :=dimension . 
aSequenceableCollection size =1 ifTrue: [value :=aSequenceableCollection first . self adjustValue . ^self ].
sorted :=self sort: aSequenceableCollection .
cut :=sorted size + 1// 2.
value :=sorted at: cut .
self adjustValue .
nextDimension :=dimension  \\(value size) +1.
cut >1 ifTrue: [left :=self class new addData: (sorted copyFrom: 1 to: cut -1 )  at: nextDimension].
right :=self class new addData: (sorted copyFrom: cut +1 to: sorted size)  at: nextDimension.! !

!KDTree methodsFor: 'private' stamp: 'wk 6/10/2013 00:45'!
adjustValue
	"necessary for IndexedKDTree "
! !

!KDTree methodsFor: 'accessing' stamp: 'WernerKassens 5/1/2011 17:29'!
farChild: aVector
^((value at: dim) < (aVector at: dim)) ifTrue: [left ] ifFalse:   [right ]! !

!KDTree methodsFor: 'accessing' stamp: 'WernerKassens 5/1/2011 17:28'!
nearChild: aVector
^((value at: dim) < (aVector at: dim)) ifTrue: [right ] ifFalse:   [left ]! !

!KDTree methodsFor: 'evaluating' stamp: 'WernerKassens 3/5/2014 10:28'!
nnSearch: aSequenceableCollection i: anInt
"search for i nearest neighbours of vector aSequenceableCollection and return the vectors "
|n|
n :=NNStore new: anInt .
self nnSearch: aSequenceableCollection near: n .
n :=n data .
^anInt =1 ifTrue: [n first ] ifFalse:   [n]  ! !

!KDTree methodsFor: 'private' stamp: 'WernerKassens 3/5/2014 10:28'!
nnSearch: aSequenceableCollection near: aNNStore
|dist|
(self nearChild: aSequenceableCollection)ifNotNilDo:[:c| c nnSearch: aSequenceableCollection near: aNNStore ].
(aNNStore maxDistance)<( (dist := (value at: dim) - (aSequenceableCollection at: dim) ) * dist ) ifTrue:[^self ].
dist := ((dist:= aSequenceableCollection - value)*dist )sum.
self add: dist to: aNNStore .
(self farChild: aSequenceableCollection)ifNotNilDo:[:c| c nnSearch: aSequenceableCollection  near: aNNStore].! !

!KDTree methodsFor: 'printing' stamp: 'wk 2/28/2014 02:20'!
printOn: aStream
      super printOn: aStream .
	aStream nextPutAll:  ' (dim: ';
	 print: dim ;
	 nextPutAll:  ' value: ';
	 print: value ;
	 nextPutAll: ' left: ';
	 print: left ;
	 nextPutAll: ' right: '; 
	 print: right ;
	 nextPut: $) .! !

!KDTree methodsFor: 'private' stamp: 'WernerKassens 3/5/2014 10:27'!
sort: aSequenceableCollection
^ aSequenceableCollection sort:  [:a :b| (a  at: dim) < (b at: dim)] ! !
